// DTS Coherent Acoustics decoder for foobar2000
// Based on dtsdec, Copyright (C) 2004 Gildas Bazin <gbazin@videolan.org>
//                            (C) 2000-2003 Michel Lespinasse <walken@zoy.org>
//                            (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
// foobar component copyright (C) 2004-2006 Janne Hyvrinen
//
// Changes:
//  0.1.7  (2006-07-22): Added support for internal cuesheets
//  0.1.6  (2006-07-12): Playback from cuesheets always started from the beginning of the file, fixed
//  0.1.5  (2006-07-09): foobar2000 v0.9.x conversion
//  0.1.4  (2005-07-24): Accepts header with reversed word order
//  0.1.3  (2005-03-28): Smarter channel order fixing. Now uses 9 channel output only for files that output to center rear speaker.
//                       Added channel order hack also to packet decoder, added configuration
//  0.1.2  (2005-03-28): Fixed crashing with invalid files
//  0.1.1  (2005-03-26): Hack to fix output channel order for odd channel modes
//  0.1    (2005-03-20): Fixed incorrect handling for files that had no LFE channel, more accurate seeking
//  0.0.9  (2005-01-01): Fixed hanging at the end of DTS files
//  0.0.8  (2004-12-27): Improved DTS-in-wav detection
//  0.0.7  (2004-10-23): Added DTS header seeking to support more DTS-in-wav files
//  0.0.6  (2004-10-20): Removed most of the changes for 0.0.5 and replaced with hacks for now
//  0.0.5  (2004-10-17): Fixes to raw DTS handling
//  0.0.4  (2004-10-15): Simplified packet decoder, added codec reporting, fixed typo in version number
//  0.0.3  (2004-10-15): Added Matroska packet decoder support

#define FD_VERSION  "0.1.7"

#include "../SDK/foobar2000.h"
using namespace pfc;
#include "../helpers/helpers.h"
//#include "resource.h"
// #include <commctrl.h>

typedef unsigned int    uint32_t;
typedef signed int      int32_t;
typedef unsigned short  uint16_t;
typedef signed short    int16_t;
typedef unsigned char   uint8_t;
typedef signed char     int8_t;

extern "C" {
#include "dts.h"
}

#define BUFFER_SIZE 24576
#define HEADER_SIZE 14
#define FRAME_SAMPLES 256

#define cfg_drc 0   // disable dynamic range compression
#define cfg_tag 1   // allow tagging
//cfg_int cfg_drc ( "drc", 0 );   // Dynamic range compression defaults to off
//cfg_int cfg_tag ( "tag", 1 );   // Write APEv2 tags by default

// -------------------------------------

static void parse_tagtype_internal(const char *p_tagtype, bool &p_have_id3v1, bool &p_have_id3v2, bool &p_have_apev2)
{
    const char *tagtype = p_tagtype;
    while(*tagtype)
    {
        unsigned delta = 0;
        while(tagtype[delta] != 0 && tagtype[delta] != '|') delta++;
        if (delta > 0)
        {
            if (!stricmp_utf8_ex(tagtype,delta,"apev1",infinite) || !stricmp_utf8_ex(tagtype,delta,"apev2",infinite))
                p_have_apev2 = true;
            else if (!stricmp_utf8_ex(tagtype,delta,"id3v1",infinite))
                p_have_id3v1 = true;
            else if (!stricmp_utf8_ex(tagtype,delta,"id3v2",infinite))
                p_have_id3v2 = true;
        }
        tagtype += delta;
        while(*tagtype == '|') tagtype++;
    }
}

class input_dts {
private:
    service_ptr_t<file> r;
    //reader *r;
    long frame_size;
    dts_state_t *state;
    t_filesize header_pos;
    t_filesize datasize;
    bool eof;

    pfc::array_t<audio_sample> output;
    uint8_t buffer[BUFFER_SIZE];
    uint8_t buf[BUFFER_SIZE];
    uint8_t *bufptr, *bufpos;

    int disable_dynrng;
    int dts_flags, nch, srate, bitrate, frame_length;
    double real_bitrate;
    unsigned int skip_samples;
    unsigned int channel_mask;
    //int hack_nch;

    __int64 skip_wav_header(abort_callback &p_abort)
    {
        t_filesize pos = r->get_position(p_abort);
        t_filesize filesize = r->get_size(p_abort);

        for (;;) {
            char temp[4];
            DWORD tmp;
            r->read ( temp, 4, p_abort );
            if ( memcmp(temp, "RIFF", 4) ) break;
            r->read ( temp, 4, p_abort );
            r->read ( temp, 4, p_abort );
            if ( memcmp(temp, "WAVE", 4) ) break;

            for (;;) {
                if ( r->read(temp, 4, p_abort) != 4 ) break;
                if ( !memcmp(temp, "fmt ", 4) ) break; //success
                if ( !memcmp(temp, "data", 4) ) break; //already got data chunk but no fmt
                if ( r->read(&tmp, 4, p_abort) != 4 ) break;
                if ( tmp & 1 ) tmp++;
                if ( tmp == 0 || r->get_position(p_abort) + tmp > filesize - 8 ) break;
                r->seek(tmp + r->get_position(p_abort), p_abort);
            }

            if ( memcmp(temp, "fmt ", 4) ) break;

            __int64 position = r->get_position(p_abort) - 4;

            if ( r->read(&tmp, 4, p_abort) != 4 ) break;
            if ( tmp < 14 || tmp > 64*1024 ) break;
            r->seek(tmp + r->get_position(p_abort), p_abort);

            char code[4];

            do {
                position += 8 + tmp + (tmp&1); //word-align all blocksizes
                r->seek(position, p_abort);
                if ( r->read(code, 4, p_abort) != 4 ) break;
                if ( r->read(temp, 4, p_abort) != 4 ) break;
            } while ( memcmp(code, "data", 4) );

            if ( memcmp(code, "data", 4) ) break;

            position += 8;
            r->seek ( position, p_abort );
            return position - pos;
        }

        r->seek ( pos, p_abort );
        return 0;
    }

    int find_dts_header ( const unsigned char *buf, unsigned int size )
    {
        unsigned int size_to_search = size-15;
        if ( size_to_search > size ) {
            // not enough data for one header
            return -1;
        }

        for ( unsigned int offset = 0; offset < size_to_search; offset++ ) {
            // sync words appear aligned in the bit stream
            if ( (buf[offset + 0] == 0x7f) && // 16 bit SYNC
                 (buf[offset + 1] == 0xfe) &&
                 (buf[offset + 2] == 0x80) &&
                 (buf[offset + 3] == 0x01) ) return offset;

            if ( (buf[offset + 0] == 0xfe) && // 16 bit SYNC, reversed word order
                 (buf[offset + 1] == 0x7f) &&
                 (buf[offset + 2] == 0x01) &&
                 (buf[offset + 3] == 0x80) ) return offset;

            if ( (buf[offset + 0] == 0xff) && // 14 bit SYNC
                 (buf[offset + 1] == 0x1f) &&
                 (buf[offset + 2] == 0x00) &&
                 (buf[offset + 3] == 0xe8) &&
                 ((buf[offset + 4] & 0xf0) == 0xf0) &&
                 (buf[offset + 5] == 0x07) ) return offset;

            if ( (buf[offset + 0] == 0x1f) && // 14 bit SYNC, reversed word order
                 (buf[offset + 1] == 0xff) &&
                 (buf[offset + 2] == 0xe8) &&
                 (buf[offset + 3] == 0x00) &&
                 (buf[offset + 5] == 0x07) &&
                 ((buf[offset + 5] & 0xf0) == 0xf0)) return offset;
        }

        return -1;
    }

    int find_first_dts_header ( unsigned int max_offset, abort_callback &p_abort )
    {
        int offset = 0;

        while (offset+sizeof(buffer) <= max_offset) {
            if ( r->read(buffer, sizeof(buffer), p_abort) != sizeof(buffer) ) return -1;
            int pos = find_dts_header ( buffer, sizeof(buffer) );
            if ( pos >= 0 ) return offset + pos;
            offset += sizeof ( buffer );
        }

        return -1;
    }

    int get_channel_count ( int flags )
    {
        int channels = 0, tmp = flags & DTS_CHANNEL_MASK;

        if ( tmp > DTS_CHANNEL_MAX ) return 0;

        if ( tmp == DTS_MONO ) channels = 1;
        else if ( tmp < DTS_3F ) channels = 2;
        else if ( tmp < DTS_3F1R ) channels = 3;
        else if ( tmp < DTS_3F2R ) channels = 4;
        else if ( tmp < DTS_4F2R ) channels = 5;
        // else if ( tmp == DTS_4F2R ) channels = 6; // not supported

        if ( channels && (flags & DTS_LFE) ) channels++;

        return channels;
    }

#if 0
    int get_output_channel_count ( int flags )
    {
        static const int output_nch[10] = {
            1,       // DTS_MONO
            2,       // DTS_CHANNEL
            2,       // DTS_STEREO
            2,       // DTS_STEREO_SUMDIFF
            2,       // DTS_STEREO_TOTAL
            3,       // DTS_3F
            9,       // DTS_2F1R
            9,       // DTS_3F1R
            6,       // DTS_2F2R
            6,       // DTS_3F2R
        };
        static const int output_nch_lfe[10] = {
            4,       // DTS_MONO
            4,       // DTS_CHANNEL
            4,       // DTS_STEREO
            4,       // DTS_STEREO_SUMDIFF
            4,       // DTS_STEREO_TOTAL
            4,       // DTS_3F
            9,       // DTS_2F1R
            9,       // DTS_3F1R
            6,       // DTS_2F2R
            6,       // DTS_3F2R
        };

        if ( flags & DTS_LFE ) {
            return output_nch_lfe[flags & DTS_CHANNEL_MASK];
        } else {
            return output_nch[flags & DTS_CHANNEL_MASK];
        }
    }
#endif

    unsigned int get_channel_map_mask ( int flags )
    {
        static const unsigned int output_nch[10] = {
            audio_chunk::channel_front_center,                                  // DTS_MONO
            audio_chunk::channel_front_left | audio_chunk::channel_front_right, // DTS_CHANNEL
            audio_chunk::channel_front_left | audio_chunk::channel_front_right, // DTS_STEREO
            audio_chunk::channel_front_left | audio_chunk::channel_front_right, // DTS_STEREO_SUMDIFF
            audio_chunk::channel_front_left | audio_chunk::channel_front_right, // DTS_STEREO_TOTAL
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center, // DTS_3F
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_back_center,  // DTS_2F1R
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center | audio_chunk::channel_back_center, // DTS_3F1R
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_back_left    | audio_chunk::channel_back_right,  // DTS_2F2R
            audio_chunk::channel_front_left | audio_chunk::channel_front_right | audio_chunk::channel_front_center | audio_chunk::channel_back_left | audio_chunk::channel_back_right, // DTS_3F2R
        };

        unsigned int map = output_nch[flags & DTS_CHANNEL_MASK];
        if (flags & DTS_LFE) map |= audio_chunk::channel_lfe;
        return map;
    }

public:
    input_dts()
    {
        state = 0;
        disable_dynrng = !cfg_drc;
    }

    ~input_dts()
    {
        if ( state ) dts_free ( state );
    }

    bool decode_can_seek() { return r->can_seek(); }
    bool decode_get_dynamic_info ( file_info &p_out, double &p_timestamp_delta ) { return false; }
    bool decode_get_dynamic_info_track ( file_info &p_out, double &p_timestamp_delta ) { return false; }
    void decode_on_idle ( abort_callback &p_abort ) { r->on_idle(p_abort); }

    static bool g_is_our_content_type ( const char *p_content_type )
    {
        return (!strcmp(p_content_type, "audio/x-dts")) || (!strcmp(p_content_type, "audio/dts"));
    }

    static bool g_is_our_path ( const char *p_path, const char *p_extension )
    {
        return stricmp_utf8(p_extension, "dts") == 0;
    }

    t_filestats get_file_stats ( abort_callback &p_abort ) { return r->get_stats(p_abort); }

    void open ( service_ptr_t<file> p_filehint, const char *p_path, t_input_open_reason p_reason, abort_callback &p_abort )
    {
        if (p_reason==input_open_info_write && !cfg_tag) throw exception_io_unsupported_format();

        r = p_filehint;
        input_open_file_helper(r, p_path, p_reason, p_abort);

        state = dts_init ( 0 );
        if ( !state ) {
            throw exception_io_data("Failed to initialize DTS decoder");
        }

        //header_pos = id3v2_calc_size ( r );
        tag_processor::skip_id3v2(r, header_pos, p_abort);
        header_pos += skip_wav_header(p_abort);

        bufptr = buf;
        bufpos = buf + HEADER_SIZE;
        nch = srate = dts_flags = frame_size = 0;
        skip_samples = 0;
        eof = false;
        //hack_nch = 0;
        channel_mask = 0;

        if ( find_first_dts_header(2*1024*1024, p_abort) < 0 ) {
            dts_free ( state );
            state = 0;
            throw exception_io_data();
        }
        r->seek ( header_pos, p_abort );

        __int64 t = find_dts_header ( buffer, sizeof(buffer) );
        if ( t < 0 ) {
            dts_free ( state );
            state = 0;
            throw exception_io_data();
        }

        frame_size = find_dts_header ( buffer+t+1, (int)(sizeof(buffer)-t-1) ) + 1;
        if ( frame_size <= 0 ) {
            dts_free ( state );
            state = 0;
            throw exception_io_data();
        }

        t = dts_syncinfo ( state, buffer+t, &dts_flags, &srate, &bitrate, &frame_length );
        if ( !t ) {
            dts_free ( state );
            state = 0;
            throw exception_io_data();
        }

        nch = get_channel_count ( dts_flags );
        if ( !srate || !bitrate || !frame_length || !nch ) {
            dts_free ( state );
            state = 0;
            throw exception_io_data();
        }

        channel_mask = get_channel_map_mask ( dts_flags );
        //hack_nch = get_output_channel_count ( dts_flags );
        real_bitrate = (double)(frame_size * 8) * ((double)srate / frame_length);
        bitrate = (int)real_bitrate;

        t = r->get_size(p_abort);
        if (t != filesize_invalid) {
            t -= header_pos;
            datasize = t - (t % frame_size);
        }
    }

    void get_info ( file_info &p_info, abort_callback &p_abort )
    {
        static const char *dts_mode_list[10] = {
            { "Mono" },
            { "Dual Mono" },
            { "Stereo" },
            { "Stereo (SUMDIFF)" },
            { "Stereo (TOTAL)" },
            { "3 front channels" },
            { "2 front, 1 rear surround channel" },
            { "3 front, 1 rear surround channel" },
            { "2 front, 2 rear surround channels" },
            { "3 front, 2 rear surround channels" }
        };

        double length = (double)datasize * 8.0 / real_bitrate;

        p_info.set_length ( length );
        p_info.info_set_int ( "samplerate", srate );
        p_info.info_set_int ( "channels", nch );
        p_info.info_set_int ( "bitrate", (int)(real_bitrate/1000) );
        p_info.info_set ( "codec", "DTS" );
        p_info.info_set ( "encoding", "lossy" );
        p_info.info_set ( "channel_mode", pfc::string_formatter() << dts_mode_list[dts_flags & DTS_CHANNEL_MASK] << (dts_flags & DTS_LFE ? " + LFE" : "") );

        t_filesize current_pos = r->get_position(p_abort);

        try {
            tag_processor::read_id3v2_trailing(r, p_info, p_abort); // read all tags from the file
        } catch(exception_io_data) { /*file had no tags */ }
        
        r->seek(current_pos, p_abort);
    }

    void decode_initialize ( unsigned p_flags, abort_callback &p_abort )
    {
        r->reopen(p_abort); //equivalent to seek to zero, except it also works on nonseekable streams
        r->seek(header_pos, p_abort);
    }

    //virtual int run ( audio_chunk *chunk )
    bool decode_run ( audio_chunk &chunk, abort_callback &p_abort )
    {
        /*
        Win32:
        0. Front Left - FL
        1. Front Right - FR
        2. Front Center - FC
        3. Low Frequency - LF
        4. Back Left - BL
        5. Back Right - BR
        6. Front Left of Center - FLC
        7. Front Right of Center - FRC
        8. Back Center - BC
        
        DTS:
        0. center
        1. left
        2. right
        3. left surround
        4. right surround
        5. LFE
        */
        static const int chan_map[10][6] = {
            { 0, 0, 0, 0, 0, 0 },       // DTS_MONO
            { 0, 1, 0, 0, 0, 0 },       // DTS_CHANNEL
            { 0, 1, 0, 0, 0, 0 },       // DTS_STEREO
            { 0, 1, 0, 0, 0, 0 },       // DTS_STEREO_SUMDIFF
            { 0, 1, 0, 0, 0, 0 },       // DTS_STEREO_TOTAL
            { 2, 0, 1, 0, 0, 0 },       // DTS_3F
            { 0, 1, 2, 0, 0, 0 },       // DTS_2F1R
            { 2, 0, 1, 3, 0, 0 },       // DTS_3F1R
            { 0, 1, 2, 3, 0, 0 },       // DTS_2F2R
            { 2, 0, 1, 3, 4, 0 },       // DTS_3F2R
        };
        static const int chan_map_lfe[10][6] = {
            { 0, 1, 0, 0, 0, 0 },       // DTS_MONO
            { 0, 1, 2, 0, 0, 0 },       // DTS_CHANNEL
            { 0, 1, 2, 0, 0, 0 },       // DTS_STEREO
            { 0, 1, 2, 0, 0, 0 },       // DTS_STEREO_SUMDIFF
            { 0, 1, 2, 0, 0, 0 },       // DTS_STEREO_TOTAL
            { 2, 0, 1, 3, 0, 0 },       // DTS_3F
            { 0, 1, 3, 2, 0, 0 },       // DTS_2F1R
            { 2, 0, 1, 4, 3, 0 },       // DTS_3F1R
            { 0, 1, 3, 4, 2, 0 },       // DTS_2F2R
            { 2, 0, 1, 4, 5, 3 },       // DTS_3F2R
        };
        /*
        static const int chan_map_hack[10][6] = {
            { 2, 3, 0, 0, 0, 0 },       // DTS_MONO
            { 0, 1, 3, 0, 0, 0 },       // DTS_CHANNEL
            { 0, 1, 3, 0, 0, 0 },       // DTS_STEREO
            { 0, 1, 3, 0, 0, 0 },       // DTS_STEREO_SUMDIFF
            { 0, 1, 3, 0, 0, 0 },       // DTS_STEREO_TOTAL
            { 2, 0, 1, 3, 0, 0 },       // DTS_3F
            { 0, 1, 8, 3, 0, 0 },       // DTS_2F1R
            { 2, 0, 1, 8, 3, 0 },       // DTS_3F1R
            { 0, 1, 4, 5, 3, 0 },       // DTS_2F2R
            { 2, 0, 1, 4, 5, 3 },       // DTS_3F2R
        };
        */

        do {
            if (eof) return false;

            t_size size = r->read(buffer, sizeof(buffer), p_abort);

            uint8_t *start = buffer;
            uint8_t *end = buffer + size;
            unsigned int samples = 0;

            while ( 1 ) {
                int len = end - start;
                if ( !len ) break;
                if ( len > bufpos - bufptr) len = bufpos - bufptr;

                memcpy ( bufptr, start, len );
                bufptr += len;
                start += len;

                if ( bufptr == bufpos ) {
                    if ( bufpos == buf + HEADER_SIZE ) {
                        int length = dts_syncinfo ( state, buf, &dts_flags, &srate, &bitrate, &frame_length );
                        if ( !length ) {
                            //console::warning ( "skip" );
                            for ( bufptr = buf; bufptr < buf + HEADER_SIZE-1; bufptr++ ) bufptr[0] = bufptr[1];
                            continue;
                        }

                        nch = get_channel_count ( dts_flags );
                        channel_mask = get_channel_map_mask ( dts_flags );
                        //hack_nch = get_output_channel_count ( dts_flags );

                        bufpos = buf + length;
                    } else {
                        int n;
                        audio_sample *out = 0;
                        level_t level = 1.0;
                        sample_t bias = 0;

                        dts_flags &= DTS_CHANNEL_MASK | DTS_LFE;
                        if ( dts_frame(state, buf, &dts_flags, &level, bias) ) goto error;
                        if ( disable_dynrng ) dts_dynrng ( state, NULL, NULL );

                        for ( n = 0; n < dts_blocks_num(state); n++ ) {
                            if ( dts_block(state) ) goto error;
                            output.grow_size ( (samples /*+ skip_samples*/ + FRAME_SAMPLES) * /*hack_*/nch );
                            out = (audio_sample *)output.get_ptr() + samples * /*hack_*/nch;
                            float *src = dts_samples ( state );
                            if ( src ) {
                                int tmp = dts_flags & (~DTS_LFE);

                                /*
                                if ( hack_nch != nch ) memset ( out, 0, sizeof(audio_sample) * FRAME_SAMPLES * hack_nch );
                                for ( int j = 0; j < nch; j++ ) {
                                    for ( int i = 0; i < FRAME_SAMPLES; i++ ) {
                                        out[i * hack_nch + chan_map_hack[tmp][j]] = src[j * FRAME_SAMPLES + i];
                                    }
                                }
                                */

                                if ( dts_flags & DTS_LFE ) {
                                    for ( int j = 0; j < nch; j++ ) {
                                        for ( int i = 0; i < FRAME_SAMPLES; i++ ) {
                                            out[i * nch + chan_map_lfe[tmp][j]] = src[j * FRAME_SAMPLES + i];
                                        }
                                    }
                                } else {
                                    for ( int j = 0; j < nch; j++ ) {
                                        for ( int i = 0; i < FRAME_SAMPLES; i++ ) {
                                            out[i * nch + chan_map[tmp][j]] = src[j * FRAME_SAMPLES + i];
                                        }
                                    }
                                }

                                samples += FRAME_SAMPLES;
                            }
                        }

                        bufptr = buf;
                        bufpos = buf + HEADER_SIZE;
                        continue;

                        error:
                        console::warning ( "error" );
                        bufptr = buf;
                        bufpos = buf + HEADER_SIZE;
                    }
                }
            }

            if ( samples > 0 ) {
                if ( samples > skip_samples ) {
                    chunk.set_data ( (audio_sample *)(output.get_ptr() + skip_samples* /*hack_*/ nch), samples - skip_samples, /*hack_*/ nch, srate, channel_mask );
                    skip_samples = 0;
                    return true;
                } else {
                    skip_samples -= samples;
                }
            } else if ( size > 0 ) { // hack to output silence with DTS in WAV (as CD players would do)
                samples = sizeof(buffer) / (2*sizeof(short));
                if ( samples > skip_samples ) {
                    output.grow_size ( (samples + skip_samples) * /*hack_*/nch );
                    memset ( (audio_sample *)(output.get_ptr() + skip_samples * /*hack_*/nch), 0, (samples-skip_samples) * /*hack_*/nch * sizeof(audio_sample) );
                    chunk.set_data ( (audio_sample *)(output.get_ptr() + skip_samples* /*hack_*/ nch), samples - skip_samples, /*hack_*/ nch, srate, channel_mask );
                    skip_samples = 0;
                    return true;
                } else {
                    skip_samples -= samples;
                }
            } else {
                return false;
            }
        } while ( 1 );
    }

    void decode_seek ( double newpos, abort_callback &p_abort )
    {
        r->ensure_seekable(); //throw exceptions if someone called decode_seek() despite of our input having reported itself as nonseekable.

        eof = false;
        skip_samples = 0;

        if ( newpos >= 0 && frame_size > 0 ) {
            __int64 offset = (__int64)((newpos * real_bitrate/8.0) + 0.5);
            offset -= offset % frame_size;
            offset += header_pos;
            if ( offset >= (__int64)(header_pos + datasize) ) {
                eof = true;
                return;
            }
            if ( offset >= (__int64)(frame_size-header_pos) ) {
                offset -= frame_size;
                skip_samples = FRAME_SAMPLES;
            }
            r->seek(offset, p_abort);

            skip_samples += (unsigned int)(((__int64)(newpos * srate)) - ((__int64)((offset-header_pos) / (real_bitrate/8.0) * srate)));
        }
    }

    void retag ( const file_info &p_info, abort_callback &p_abort )
    {
        if (!cfg_tag) throw exception_io_unsupported_format(); // exit out when tags are not wanted
        r->ensure_seekable();

        const char *tagtype = p_info.info_get("tagtype");
        if (tagtype == 0) tagtype = "";

        bool id3v1=false, id3v2=false, apev2=false;
        parse_tagtype_internal(tagtype, id3v1, id3v2, apev2);

        if (!id3v2 && !apev2) apev2 = true;

        tag_processor::write_multi(r, p_info, p_abort, id3v1, id3v2, apev2);
        //tag_processor::remove_id3v2_trailing(r, p_abort);
        //tag_processor::write_apev2(r, p_info, p_abort);
    }
};

//static input_singletrack_factory_t<input_dts> g_input_dts_factory;
static input_cuesheet_factory_t<input_dts> g_input_dts_factory;

DECLARE_COMPONENT_VERSION("DTS decoder", FD_VERSION,
  "DTS decoding powered by libdca v0.0.2 by Gildas Bazin.\n"
  "Homepage for libdca: http://developers.videolan.org/libdca.html\n"
  "\n"
  "foobar component by Janne Hyvärinen.\n"
  "Licensed under GNU GPL.\n");

DECLARE_FILE_TYPE("DTS files", "*.DTS");
